-- | Constant-Pool entries.

{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DataKinds #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE OrPatterns #-}
module Language.Java.Classfile.ConstantPool.Entry (Entry(..), StorageCount(..), storageCount, MethodHandleInfo(..)) where
import GHC.Generics (Generic, Generically(..))
import Language.Java.Classfile.Extractable (Extractable (extract))
import Language.Java.Classfile.Extractable.WithTag (Word8Tag)
import Language.Java.Classfile.ConstantPool.References (ClassReference, NameAndTypeReference, MethodHandleReferenceKind, OpaqueReference, BootstrapMethodIndex, Utf8Reference)
import Data.Int (Int32, Int64)
import Data.Word (Word16)
import Data.Text (Text)
import Language.Java.Classfile.Extract (Extract, bytes)
import qualified Data.Text.Encoding as Text
import qualified Data.ByteString.Lazy as ByteString
import qualified Data.ByteString as StrictByteString

-- | A single entry. Double and Long are not followed by something unusable, they are duplicated instead.

data Entry
  = Utf8                (Word8Tag 1  (SizedText Word16))
  -- ^ Java-modified Utf8
  | Integer             (Word8Tag 3  Int32)
  -- ^ Constant value for ldc
  | Float               (Word8Tag 4  Float)
  -- ^ Constant value for ldc
  | Long                (Word8Tag 5  Int64)
  -- ^ Constant value for ldc
  | Double              (Word8Tag 6  Double)
  -- ^ Constant value for ldc
  | Class               (Word8Tag 7  Utf8Reference)
  -- ^ Reference to another class: full name
  | String              (Word8Tag 8  Utf8Reference)
  -- ^ Constant value for ldc or ConstantValue attributes.
  | FieldRef            (Word8Tag 9  (ClassReference, NameAndTypeReference))
  -- ^ Reference to a field: (containing class, descriptor)
  | MethodRef           (Word8Tag 10 (ClassReference, NameAndTypeReference))
  -- ^ Reference to a method: (containing class, descriptor)
  | InterfaceMethodRef  (Word8Tag 11 (ClassReference, NameAndTypeReference))
  -- ^ Reference to an interface method: (containing class, descriptor)
  | NameAndType         (Word8Tag 12 (Utf8Reference, Utf8Reference))
  -- ^ Name and Type: (name, type)
  | MethodHandle        (Word8Tag 15 MethodHandleInfo)
  -- ^ Reference to a Method, but more *dynamic*
  | MethodType          (Word8Tag 16 Utf8Reference)
  -- ^ Descriptor of a Method
  | Dynamic             (Word8Tag 17 (BootstrapMethodIndex, NameAndTypeReference))
  -- ^ Unspeakable horrors: index into the bootstrapmethod attribute array, NameAndType must refer to a field
  | InvokeDynamic       (Word8Tag 18 (BootstrapMethodIndex, NameAndTypeReference))
  -- ^ Unspeakable horrors: index into the bootstrapmethod attribute array, NameAndType must refer to a method
  | Module              (Word8Tag 19 Utf8Reference)
  -- ^ Some module description
  | Package             (Word8Tag 20 Utf8Reference)
  -- ^ Some package description
  deriving stock (Show, Generic)
  deriving Extractable via Generically Entry

-- | Holds the invariants of MethodHandles (Only certain references are allowed after some kinds)
--
-- TODO: Implement the invariants

data MethodHandleInfo = MethodHandleInfo MethodHandleReferenceKind OpaqueReference
  deriving stock (Show, Generic)
  deriving Extractable via Generically MethodHandleInfo

-- | Extractor newtype for a java-utf-text with size tag specified as a type argument.

newtype SizedText sizeType = SizedText Text
  deriving stock (Show)

instance (Integral sizeType, Extractable sizeType) => Extractable (SizedText sizeType) where
  extract :: Extract (SizedText sizeType)
  extract = do
    size <- extract @sizeType
    slice <- StrictByteString.concat . ByteString.toChunks <$> bytes (fromIntegral size) -- TODO: Replace java zeroes with normal zeroes, normalize 2x3 bytes into 4 byte utf8

    case Text.decodeUtf8' slice of
      Left err -> fail $ show err
      Right t -> pure $ SizedText t

-- | How many indices does a 'Entry' take?

data StorageCount = Once | Twice

-- | Get the count for a specific entry.

storageCount :: Entry -> StorageCount
storageCount = \case
  (Double _ ; Long _) -> Twice
  _  -> Once
