-- | Extract a Classfile value out of 'ByteString' input.

{-# LANGUAGE DeriveFunctor #-}
{-# LANGUAGE InstanceSigs #-}
module Language.Java.Classfile.Extract (Extract(), bytes, runExtract, expectRaw, expectEqual, traceType, traceConstructor, traceField, traceIndex, Reason(..), Trace(..), Expected(..), Actual(..), TypeName(..)) where

import Data.ByteString.Lazy (ByteString)
import qualified Data.ByteString.Lazy as ByteString
import Control.Applicative (Alternative (empty, (<|>)))
import Data.Text (Text)
import qualified Data.Text as Text
import Data.Typeable (Typeable, typeOf)

-- | Extractor Monad. Computations running in this monad will automatically keep track of used resources and backtrack arbitrarily.

newtype Extract a = Extract (Continuation a)
  deriving (Functor)

-- | Functions work wonders when defining monads.

type Continuation a = ByteString -> Reply a

-- | Result used to abort computations on failure.

data Reply a
  = Done ByteString a -- rest, result
  | Fail [Trace] Reason
  deriving (Functor)

-- | Type Alias for the Show instance
newtype Expected = Expected Text
  deriving Show

-- | Type Alias for the Show instance
newtype Actual = Actual Text
  deriving Show

-- | Type Alias for the Show instance
newtype TypeName = TypeName Text
  deriving Show

-- | Why did the computation fail?

data Reason
  = EndOfInput
  -- ^ The parser expected more input bytes
  | UnexpectedValue Expected Actual TypeName
  -- ^ A specific value was expected but not provided
  | Custom Text
  -- ^ Something else went wrong
  | Unknown
  -- ^ Someone used the 'empty' function from 'Alternative'
  deriving Show

-- | Where did the computation fail? Calls to these functions are auto-generated by the Generic instances.

data Trace
  = InType Text
  | InConstructor Text
  | InField Text
  | AtIndex Word
  deriving Show

instance Applicative Extract where
  -- | Don't consume any input
  pure :: a -> Extract a
  pure x = Extract $ \ rest -> Done rest x

  -- | One by one, Fail short-circuits it all.
  (<*>) :: Extract (a -> b) -> Extract a -> Extract b
  (<*>) (Extract computeF) (Extract computeX) = Extract $ \ input -> case computeF input of
    Done rest f -> case computeX rest of
      Done rest' x -> Done rest' (f x)
      Fail ts r -> Fail ts r
    Fail ts r -> Fail ts r

instance Monad Extract where
  -- | One by one, Fail short-circuits it all.
  (>>=) :: Extract a -> (a -> Extract b) -> Extract b
  (>>=) (Extract computeA) f = Extract $ \ input -> case computeA input of
    Done rest a -> let
      (Extract computeB) = f a
      in computeB rest
    Fail ts r -> Fail ts r

instance Alternative Extract where
  -- | Fail the computation.
  empty :: Extract a
  empty = Extract $ const (Fail [] Unknown)

  -- | This implementation is left-biased. Only when the left computation fails will it run the right one.
  (<|>) :: Extract a -> Extract a -> Extract a
  (<|>) (Extract left) (Extract right) = Extract $ \ input -> case left input of
    Fail _ _ -> right input
    t -> t

instance MonadFail Extract where
  -- | Produces the 'Custom' Failure Reason
  fail :: String -> Extract a
  fail = Extract . const . Fail [] . Custom . Text.pack

-- | Get a specified count of bytes. Fail if there are not enough bytes available.

bytes :: Word -> Extract ByteString
bytes count = Extract $ \ input -> let
  count' = fromIntegral count
  (bs, rest) = ByteString.splitAt count' input
  in if ByteString.length bs /= count'
    then Fail [] EndOfInput
    else Done rest bs

-- | Fail if the specified bytestring is not immediately parseable

expectRaw :: ByteString -> Extract ByteString
expectRaw e = Extract $ \ input -> let
     (actual, rest) = ByteString.splitAt (ByteString.length e) input
     in case actual == e of
       True -> Done rest actual
       False -> Fail [] (mismatchReason e actual)

-- | Fail if the Extractor doesn't yield the specified value

expectEqual :: (Eq a, Show a, Typeable a) => Extract a -> a -> Extract a
expectEqual (Extract computeActual) expected = Extract $ \ input -> case computeActual input of
  Done rest actual -> if actual == expected
    then Done rest actual
    else Fail [] (mismatchReason expected actual)
  failed -> failed

-- | Constructs the reason, arguments are in this order: expected, actual

mismatchReason :: (Typeable a, Show a) => a -> a -> Reason
mismatchReason expected actual = UnexpectedValue (Expected expectedText) (Actual actualText) (TypeName typeName)
  where
    actualText = textShow actual
    expectedText = textShow expected
    typeName = textShow $ typeOf expected

-- | You know what this is

textShow :: Show a => a -> Text
textShow = Text.pack . show

-- | Apply the extractor to the bytestring, returns the result and the rest on success, otherwise a best-effort traceback.

runExtract :: ByteString -> Extract b -> Either (Reason, [Trace]) (ByteString, b)
runExtract string (Extract computation) = case computation string of
  Fail trace reason -> Left (reason, trace)
  Done rest x -> Right (rest, x)

-- | Trace the contained extractor, the backtrace will supply a hint with the given typeName

traceType :: Text -> Extract a -> Extract a
traceType typeName (Extract computeA) = Extract $ \ input -> case computeA input of
  Fail ts r -> Fail (InType typeName:ts) r
  t -> t

-- | Trace the contained extractor, the backtrace will supply a hint with the given constructor name

traceConstructor :: Text -> Extract a -> Extract a
traceConstructor conName (Extract computeA) = Extract $ \ input -> case computeA input of
  Fail ts r -> Fail (InConstructor conName:ts) r
  t -> t

-- | Trace the contained extractor, the backtrace will supply a hint with the given Field name

traceField :: Text -> Extract a -> Extract a
traceField fieldName (Extract computeA) = Extract $ \ input -> case computeA input of
  Fail ts r -> Fail (InField fieldName:ts) r
  t -> t

-- | Trace the contained extractor, the backtrace will indicate that the computation was at a specific index.

traceIndex :: Word -> Extract a -> Extract a
traceIndex index (Extract computeA) = Extract $ \ input -> case computeA input of
  Fail ts r -> Fail (AtIndex index:ts) r
  t -> t

